package io.zero88.jpa;

import java.io.IOException;
import java.io.Serializable;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import io.zero88.jpa.Order.OrderSerializer;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;

import lombok.AccessLevel;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.ToString;
import lombok.experimental.Accessors;

/**
 * Represents for Order.
 *
 * @since 1.0.0
 */
@Getter
@Accessors(fluent = true)
@ToString
@EqualsAndHashCode
@RequiredArgsConstructor(access = AccessLevel.PRIVATE)
@JsonSerialize(using = OrderSerializer.class)
public final class Order implements Serializable {

    @NonNull
    private final String property;
    @NonNull
    private final Direction direction;

    /**
     * By asc order.
     *
     * @param property the property
     * @return the order
     * @since 1.0.0
     */
    public static Order byASC(@NonNull String property) {
        return by(property, Direction.ASC);
    }

    /**
     * By desc order.
     *
     * @param property the property
     * @return the order
     * @since 1.0.0
     */
    public static Order byDESC(@NonNull String property) {
        return by(property, Direction.DESC);
    }

    /**
     * By order.
     *
     * @param property  the property
     * @param direction the direction
     * @return the order
     * @since 1.0.0
     */
    public static Order by(@NonNull String property, Direction direction) {
        return new Order(property, Optional.ofNullable(direction).orElse(Direction.ASC));
    }

    /**
     * By order.
     *
     * @param property  the property
     * @param direction the direction
     * @return the order
     * @since 1.0.0
     */
    public static Order by(@NonNull @JsonProperty("property") String property,
        @JsonProperty("direction") String direction) {
        return by(property, Direction.parse(direction));
    }

    @JsonCreator
    private static Order by(@NonNull Map<String, String> properties) {
        final Entry<String, String> val = properties.entrySet()
                                                    .stream()
                                                    .findFirst()
                                                    .orElseThrow(() -> new RuntimeException("Empty value"));
        return by(val.getKey(), Direction.parse(val.getValue()));
    }

    public static class OrderSerializer extends JsonSerializer<Order> {

        @Override
        public void serialize(Order value, JsonGenerator gen, SerializerProvider serializers) throws IOException {
            gen.writeStartObject();
            gen.writeObjectField(value.property(), value.direction());
            gen.writeEndObject();
        }

    }

}
